package dk.nsp.cbs.perf.runner;

import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.List;
import java.util.Properties;

import dk.nsp.cbs.perf.config.CredentialVaultInfo;
import dk.nsp.cbs.perf.config.NasAccess;
import dk.nsp.cbs.perf.config.NasConfiguration;
import dk.nsp.cbs.perf.nas.DGWSService;
import dk.nsp.cbs.perf.nas.NasService;
import dk.nsp.cbs.perf.nas.Subscription;
import dk.nsp.cbs.perf.service.CbsClient;
import dk.nsp.cbs.perf.http.Response;
import dk.nsp.cbs.perf.service.StsJwtClient;
import dk.nsp.cbs.perf.service.vault.VaultService;
import dk.sosi.seal.model.IdentityToken;
import dk.sosi.seal.pki.Federation;
import dk.sosi.seal.pki.SOSITestFederation;
import dk.sosi.seal.vault.CredentialVault;
import org.apache.http.client.HttpClient;
import org.apache.http.impl.client.DefaultHttpClient;
import org.cfg4j.provider.ConfigurationProvider;
import org.cfg4j.provider.ConfigurationProviderBuilder;
import org.cfg4j.source.context.environment.ImmutableEnvironment;
import org.cfg4j.source.files.FilesConfigurationSource;

public class Runner {
  public static void main(String[] args) {
    // Configuration; http://www.cfg4j.org/releases/latest/#users-guide

    // etc/local is running against a locally started stub server
    final String envName = System.getProperty("perf.env", "etc/local"); // etc/dev is also possible for testing against the external test setup
    final ConfigurationProvider cfgProvider = new ConfigurationProviderBuilder()
      .withConfigurationSource(new FilesConfigurationSource()) // we can also merge configuration files
      .withEnvironment(new ImmutableEnvironment(envName))
      .build();

    final String vaultDirStr = cfgProvider.getProperty("vaultDir", String.class);
    final CredentialVaultInfo idpVaultInfo = cfgProvider.bind("idpVault", CredentialVaultInfo.class);
    final CredentialVaultInfo signingVaultInfo = cfgProvider.bind("signingVault", CredentialVaultInfo.class);

    final String stsJwtUrl = cfgProvider.getProperty("sts.jwt.location", String.class);
    final String stsDgwsUrl = cfgProvider.getProperty("sts.dgws.location", String.class);
    final String cbsUrl = cfgProvider.getProperty("cbs.location", String.class);

    final String stsJwtAuduence = cfgProvider.getProperty("sts.jwt.audience", String.class);
    final String stsJwtKid = cfgProvider.getProperty("sts.jwt.kid", String.class);

    final NasConfiguration nasConfiguration = cfgProvider.bind("nas", NasConfiguration.class);
    final NasAccess nasAccess = cfgProvider.bind("nasAccess", NasAccess.class);
    final CredentialVaultInfo nasAccessVaultInfo = cfgProvider.bind("nasAccess.vault", CredentialVaultInfo.class);

    final String nasTopic = cfgProvider.getProperty("nas.topic", String.class);

    // For testing configuration setup
//    if (true)
//      return;

    // Setup

    final Path vaultDir = Paths.get(vaultDirStr);
    final HttpClient httpClient = new DefaultHttpClient();
    final VaultService vaultService = new VaultService(vaultDir::resolve);

    final CredentialVault idpVault = vaultService.getVaultFromResource(idpVaultInfo);
    final CredentialVault signingVault = vaultService.getVaultFromResource(signingVaultInfo);
    final CredentialVault nasVault = vaultService.getVaultFromResource(nasAccessVaultInfo);

    final Federation federation = new SOSITestFederation(new Properties());

    final StsJwtClient stsJwtClient = new StsJwtClient(stsJwtUrl,
                                                       httpClient,
                                                       stsJwtAuduence,
                                                       stsJwtKid,
                                                       idpVault,
                                                       signingVault);

    final CbsClient cbsClient = new CbsClient(cbsUrl, httpClient, federation, signingVault);

    final NasService nasService = new NasService(nasConfiguration,
                                                 new DGWSService(stsDgwsUrl,
                                                                 nasAccess,
                                                                 nasVault,
                                                                 httpClient,
                                                                 federation),
                                                 httpClient);

    // final Subscription subscription = nasService.subscribe(nasTopic);
    // final List<String> strings = nasService.pullJobIds(subscription);

    new TestRunner(stsJwtClient, cbsClient).run();
  }

  public static class TestRunner {

    private final StsJwtClient stsJwtClient;
    private final CbsClient cbsClient;

    public TestRunner(StsJwtClient stsJwtClient, CbsClient cbsClient) {
      this.stsJwtClient = stsJwtClient;
      this.cbsClient = cbsClient;
    }

    public void run() {
      final IdentityToken identityToken
        = stsJwtClient.call("0501792275",
                            "SubjectDN={SERIALNUMBER=PID:9208-2002-2-514358910503 + CN=Lars Larsen, O=Ingen organisatorisk tilknytning, C=DK},IssuerDN={CN=TRUST2408 Systemtest XIX CA, O=TRUST2408, C=DK},CertSerial={1393564623}");

      final CbsClient.CbsResponse response = cbsClient.call(identityToken, "0101010101");
      System.out.println("Got a hit: " + response.isHit() + "[jobId: " + response.getJobId() + "]");
    }
  }
}
