package dk.nsp.cbs.perf.runner;

import dk.nsp.cbs.perf.config.CredentialVaultInfo;
import dk.nsp.cbs.perf.config.NasAccess;
import dk.nsp.cbs.perf.config.NasConfiguration;
import dk.nsp.cbs.perf.config.TestDriverConfig;
import dk.nsp.cbs.perf.nas.DGWSService;
import dk.nsp.cbs.perf.nas.NasService;
import dk.nsp.cbs.perf.nas.Subscription;
import dk.nsp.cbs.perf.service.CbsClient;
import dk.nsp.cbs.perf.service.StsJwtClient;
import dk.nsp.cbs.perf.service.vault.VaultService;
import dk.sosi.seal.pki.Federation;
import dk.sosi.seal.pki.SOSITestFederation;
import dk.sosi.seal.vault.CredentialVault;
import org.apache.commons.httpclient.MultiThreadedHttpConnectionManager;
import org.apache.http.HttpHost;
import org.apache.http.client.HttpClient;
import org.apache.http.conn.ClientConnectionManager;
import org.apache.http.conn.routing.HttpRoute;
import org.apache.http.impl.client.DefaultHttpClient;
import org.apache.http.impl.conn.PoolingClientConnectionManager;
import org.apache.log4j.xml.DOMConfigurator;
import org.cfg4j.provider.ConfigurationProvider;
import org.cfg4j.provider.ConfigurationProviderBuilder;
import org.cfg4j.source.context.environment.ImmutableEnvironment;
import org.cfg4j.source.files.FilesConfigurationSource;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.IOException;
import java.lang.reflect.GenericDeclaration;
import java.net.URL;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.Objects;
import java.util.Properties;
import java.util.concurrent.LinkedTransferQueue;
import java.util.concurrent.ThreadLocalRandom;

/**
 * @author jns, Arosii Information Systems A/S
 *
 * Get some personnumre from SDS database:
 * select distinct CPR from Person order by rand() limit 1000 into outfile 'rand-cpr.txt' lines terminated by '\n';
 *
 * Optional system properties:
 * <ul>
 *   <li>test.http.connman.max.total (default 200) - denotes the maximum number of http connections across all endpoints</li>
 *   <li>test.http.connman.stsjwt.max (default 95) - denotes the maximum number of http connections to the STS JWT2IDWS endpoint</li>
 *   <li>test.http.connman.cbs.max (default 95) - denotes the maximum number of http connections to the CBS endpoint</li>
 *   <li>test.http.connman.stsdgws.max (default 10) - denotes the maximum number of http connections to the STS DGWS signing endpoint</li>
 * </ul>
 */
public class TestDriver {

  public static void main(String[] args) throws IOException, InterruptedException {
    // Configuration; http://www.cfg4j.org/releases/latest/#users-guide
    final String envName = System.getProperty("perf.env", ".");
    final ConfigurationProvider cfgProvider = new ConfigurationProviderBuilder()
        .withConfigurationSource(new FilesConfigurationSource()) // we can also merge configuration files
        .withEnvironment(new ImmutableEnvironment(envName))
        .build();

    final String vaultDirStr = cfgProvider.getProperty("vaultDir", String.class);
    final CredentialVaultInfo idpVaultInfo = cfgProvider.bind("idpVault", CredentialVaultInfo.class);
    final CredentialVaultInfo signingVaultInfo = cfgProvider.bind("signingVault", CredentialVaultInfo.class);

    final String stsJwtUrl = cfgProvider.getProperty("sts.jwt.location", String.class);
    final String stsDgwsUrl = cfgProvider.getProperty("sts.dgws.location", String.class);
    final String cbsUrl = cfgProvider.getProperty("cbs.location", String.class);

    final String stsJwtAuduence = cfgProvider.getProperty("sts.jwt.audience", String.class);
    final String stsJwtKid = cfgProvider.getProperty("sts.jwt.kid", String.class);

    final NasConfiguration nasConfiguration = cfgProvider.bind("nas", NasConfiguration.class);
    final NasAccess nasAccess = cfgProvider.bind("nasAccess", NasAccess.class);
    final CredentialVaultInfo nasAccessVaultInfo = cfgProvider.bind("nasAccess.vault", CredentialVaultInfo.class);

    final String nasTopic = cfgProvider.getProperty("nas.topic", String.class);

    final TestDriverConfig testDriverConfig = cfgProvider.bind("drv", TestDriverConfig.class);

    final Path vaultDir = Paths.get(vaultDirStr);

    final PoolingClientConnectionManager connectionManager = new PoolingClientConnectionManager();
    connectionManager.setMaxTotal(Integer.getInteger("test.http.connman.max.total", 200));
    connectionManager.setMaxPerRoute(new HttpRoute(new HttpHost(new URL(stsJwtUrl).getHost())),Integer.getInteger("test.http.connman.stsjwt.max", 95));
    connectionManager.setMaxPerRoute(new HttpRoute(new HttpHost(new URL(cbsUrl).getHost())),Integer.getInteger("test.http.connman.cbs.max", 95));
    connectionManager.setMaxPerRoute(new HttpRoute(new HttpHost(new URL(stsDgwsUrl).getHost())),Integer.getInteger("test.http.connman.stsdgws.max", 10));
    final HttpClient httpClient = new DefaultHttpClient(connectionManager);

    final VaultService vaultService = new VaultService(vaultDir::resolve);

    final CredentialVault idpVault = vaultService.getVaultFromResource(idpVaultInfo);
    final CredentialVault signingVault = vaultService.getVaultFromResource(signingVaultInfo);
    final CredentialVault nasVault = vaultService.getVaultFromResource(nasAccessVaultInfo);

    final Federation federation = new SOSITestFederation(new Properties());

    final StsJwtClient stsJwtClient = new StsJwtClient(stsJwtUrl,
        httpClient,
        stsJwtAuduence,
        stsJwtKid,
        idpVault,
        signingVault);

    final CbsClient cbsClient = new CbsClient(cbsUrl, httpClient, federation, signingVault);

    final NasService nasService = new NasService(nasConfiguration,
        new DGWSService(stsDgwsUrl,
            nasAccess,
            nasVault,
            httpClient,
            federation),
        httpClient);

    final Subscription subscription = nasService.subscribe(nasConfiguration.topic());

    new RampUp(cbsClient, stsJwtClient, nasService, subscription, testDriverConfig).execute();
    if (!testDriverConfig.rampUpOnly()) {
      new Conductor(cbsClient, stsJwtClient, nasService, subscription, testDriverConfig).play();
    }
  }

  public static class InData {
    private final Path p = Paths.get(System.getProperty("perf.env"), "cpr.txt");

    public InData() {
      if (!Files.exists(p)) {
        System.err.println("cpr.txt does not exist.");
        System.exit(-1);
      }
    }

    public Path path() {
      return p;
    }
  }

  public static class Distribution {
    private final double hitWeight;

    public Distribution(final String dist) {
      final double hitPerc = Double.parseDouble(dist.split(",")[0]);
      final double missPerc = Double.parseDouble(dist.split(",")[1]);
      if ((hitPerc + missPerc) != 100) {
        System.err.println("Please configure for a 100 % distribution.");
        System.exit(-1);
      }

      if (hitPerc > missPerc) {
        hitWeight = 1.0d / (hitPerc / missPerc);
      } else if (hitPerc < missPerc) {
        hitWeight = 1.0d - (hitPerc / missPerc);
      } else {
        hitWeight = 0.5d;
      }
    }

    public boolean putInCache() {
      return ThreadLocalRandom.current().nextDouble() >= hitWeight;
    }
  }

  public static class TransitElement {
    final String jobId;
    final String pnr;
    public TransitElement(String jobId, String pnr) {
      this.jobId = jobId;
      this.pnr = pnr;
    }

    public String getJobId() {
      return jobId;
    }

    @Override
    public boolean equals(Object o) {
      if (this == o) return true;
      if (o == null || getClass() != o.getClass()) return false;
      TransitElement that = (TransitElement) o;
      return jobId.equals(that.jobId) &&
          pnr.equals(that.pnr);
    }

    @Override
    public int hashCode() {
      return Objects.hash(jobId, pnr);
    }
  }
}